// TSphere.h: interface for the TSphere class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TSPHERE_H__2748A7F9_EA12_11D3_A68A_000000000000__INCLUDED_)
#define AFX_TSPHERE_H__2748A7F9_EA12_11D3_A68A_000000000000__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//#include <windows.h>
#include <GL/gl.h>
#include "TBaseShape.h"
#include "global.h"


#pragma warning(disable: 4244)	//Shut up about double to sreal casting
#pragma warning(disable: 4305)	//'=' : truncation from 'const double' to 'sreal'




//PURPOSE:	This class is capable of doing a sphere and ellipsoids (3 radius values)
//			Surface rendering: method flat or gouraud, wireframe (2 kinds)
//			Calculates normals for flat or gouraud.
//			Can render bumblebee or checker, 2 colors.
//			Can generate texture coordinates for OpenGL, and can render
//			textured in a single shot.
//			Texturing binding, textures, textureIDs must be handled by the application.
class TSphere : public TBaseShape
{
public:
	TSphere();
	virtual ~TSphere();

	//**** Startup functions
	sint StartupByPolygon(sreal *centerXYZ, sreal *orientation, sreal radius, sint polygonNumber, tbool isGenerateIndex);
	sint StartupByABCPolygon(sreal *centerXYZ, sreal *orientation, sreal aRadius, sreal bRadius, sreal cRadius, sint polygonNumber, tbool isGenerateIndex);
	sint StartupByStacksAndSlices(sreal *centerXYZ, sreal *orientation, sreal radius, sint stacks, sint slices, tbool isGenerateIndex);
	sint StartupByABCStacksAndSlices(sreal *centerXYZ, sreal *orientation, sreal aRadius, sreal bRadius, sreal cRadius, sint stacks, sint slices, tbool isGenerateIndex);
	sint StartupByABCStacksAndSlices_StackBound(sreal *centerXYZ, sreal *orientation, sreal aRadius, sreal bRadius, sreal cRadius, sint stacks, sint slices, tbool isGenerateIndex,
		sint startingStack, sint endingStack);

	sint StartupByABCStacksAndSlices2(tbool isGenerateIndex, INDEXFORMAT indexPrecision, GLHVERTEXFORMAT vertexFormat,
		sreal *centerXYZ, sreal *orientation, sreal aRadius, sreal bRadius, sreal cRadius, sint stacks, sint slices,
		sint texCoordStyle, sreal scaleFactorS, sreal scaleFactorT);

	//**** Deallocators 
	void DeallocateAllMemory();

protected:
	void GeneralReset();
	sint SameStartupPatternForSphere(sreal *centerXYZ, sreal *orientation, tbool isGenerateIndex);
	sint SameStartupPatternForSphere2(sreal *centerXYZ, sreal *orientation);

public:
	sint GenerateVertices();
	sint GenerateIndices();
	sint GenerateNormals();
	sint GenerateTexCoords(sint texUnit, sint numberOfComponents, sint isGlobeMapOn, sint invertS, sint invertT, sint texturingMode, sint chooseVertex, sreal orientationAngle, sreal widthTextureLimit, sreal heightTextureLimit, sreal sScale, sreal tScale);
	sint GenerateTangentAndBinormalVectors();

	sint GenerateVertices2();
	sint GenerateTexCoords2(sint invertS, sint invertT);

	//sint UpdateRadius(sreal radius);
	//sint UpdateRadiusABC(sreal aRadius, sreal bRadius, sreal cRadius);
	//sint UpdatePolygonNumber(sint polygonNumber);
	//sint UpdateStacksAndSlices(sint stacks, sint slices);
	
	//Getter methods
	sint GetPolygonMinimum() { return 9;}
	sint GetStacksMinimum() { return 3;}
	sint GetSlicesMinimum() { return 3;}
	sint GetStacksNumber() { return Stacks;}
	sint GetSlicesNumber() { return Slices;}
	sint GetPolygonNumber() { return PolygonNumber;}
	sreal GetRadiusA() { return RadiusA;}
	sreal GetRadiusB() { return RadiusB;}
	sreal GetRadiusC() { return RadiusC;}



	//**** Variables
	sreal	RadiusA, RadiusB, RadiusC;
	sint	Stacks;												//Number of divisions in the Theta direction
	sint	StartingStack, EndingStack;
	sint	Slices;												//Number of divisions in the Phi direction

	//sint	StartIndex, EndIndex;								//For glDrawRangeElements

	
};



#endif // !defined(AFX_TSPHERE_H__2748A7F9_EA12_11D3_A68A_000000000000__INCLUDED_)

//LOG:
//Thursday, March 6, 2000
//Updating TSphere class
//Looking into phong rendering problem
//
//Friday, August 25, 2000
//Reworking GL_SELECT related functions
//Added TranslateCenterTo() function
//
//Monday, January 29, 2001
//Updating along with TCylinder
//
//Saturday, June 16, 2001
//Added swap color functions
//
//Thursday, August 15, 2002
//Updating for primitivelist support
//Only gouraud shading is posssible now.
//
//Saturday, March 19, 2005
//It's time to bring this up to date with new programming style
//and deriving off of TBaseShape class.
//
//Monday, March 21, 2005
//Continuing. Deleting some of the old crap. I won't use my "primitive list" feature in TSphere.
//
//Thursday, Sept 28, 2006
//I needed the ability to create dome (half-sphere, so why add the ability to generate
//the sphere between stack X and stack Y
//Adding StartupByABCStacksAndSlices_StackBound
//
//Saturday, Feb 24, 2007
//Adding support for interlaced vertex format :
//SameStartupPatternForSphere2
//GenerateVertices2
//GenerateTexCoords2
//




